from . import base


class TestBasicTypes(base.SchemaNodeTestCase):

    def test_no_object(self):
        self.assertResult({})

    def test_string(self):
        self.add_object("string")
        self.assertResult({"type": "string"})

    def test_integer(self):
        self.add_object(1)
        self.assertResult({"type": "integer"})

    def test_number(self):
        self.add_object(1.1)
        self.assertResult({"type": "number"})

    def test_boolean(self):
        self.add_object(True)
        self.assertResult({"type": "boolean"})

    def test_null(self):
        self.add_object(None)
        self.assertResult({"type": "null"})


class TestArrayList(base.SchemaNodeTestCase):

    def setUp(self):
        base.SchemaNodeTestCase.setUp(self)

    def test_empty(self):
        self.add_object([])
        self.assertResult({"type": "array"})

    def test_monotype(self):
        self.add_object(["spam", "spam", "spam", "eggs", "spam"])
        self.assertResult({"type": "array", "items": {"type": "string"}})

    def test_multitype(self):
        self.add_object([1, "2", None, False])
        self.assertResult({
            "type": "array",
            "items": {
                "type": ["boolean", "integer", "null", "string"]}
        })
        self.assertObjectValidates([False, None, "2", 1])

    def test_nested(self):
        self.add_object([
            ["surprise"],
            ["fear", "surprise"],
            ["fear", "surprise", "ruthless efficiency"],
            ["fear", "surprise", "ruthless efficiency",
             "an almost fanatical devotion to the Pope"]
        ])
        self.assertResult({
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "string"}}
        })


class TestArrayTuple(base.SchemaNodeTestCase):

    def setUp(self):
        base.SchemaNodeTestCase.setUp(self)

    def test_empty(self):
        self.add_schema({"type": "array", "items": []})

        self.add_object([])
        self.assertResult({"type": "array", "items": [{}]})

    def test_empty_schema(self):
        self.add_schema({"type": "array", "items": [{}]})

        self.add_object([])
        self.assertResult({"type": "array", "items": [{}]})

    def test_multitype(self):
        self.add_schema({"type": "array", "items": []})

        self.add_object([1, "2", "3", None, False])

        self.assertResult({
            "type": "array",
            "items": [
                {"type": "integer"},
                {"type": "string"},
                {"type": "string"},
                {"type": "null"},
                {"type": "boolean"}]
        })
        self.assertObjectDoesNotValidate([1, 2, "3", None, False])

    def test_nested(self):
        self.add_schema(
            {"type": "array", "items": {"type": "array", "items": []}})

        self.add_object([
            ["surprise"],
            ["fear", "surprise"],
            ["fear", "surprise", "ruthless efficiency"],
            ["fear", "surprise", "ruthless efficiency",
             "an almost fanatical devotion to the Pope"]
        ])
        self.assertResult({
            "type": "array",
            "items": {
                "type": "array",
                "items": [
                    {"type": "string"},
                    {"type": "string"},
                    {"type": "string"},
                    {"type": "string"}
                ]
            }
        })


class TestObject(base.SchemaNodeTestCase):

    def test_empty_object(self):
        self.add_object({})
        self.assertResult({"type": "object"})

    def test_basic_object(self):
        self.add_object({
            "Red Windsor": "Normally, but today the van broke down.",
            "Stilton": "Sorry.",
            "Gruyere": False})
        self.assertResult({
            "required": ["Gruyere", "Red Windsor", "Stilton"],
            "type": "object",
            "properties": {
                "Red Windsor": {"type": "string"},
                "Gruyere": {"type": "boolean"},
                "Stilton": {"type": "string"}
            }
        })


class TestComplex(base.SchemaNodeTestCase):

    def test_array_in_object(self):
        self.add_object({"a": "b", "c": [1, 2, 3]})
        self.assertResult({
            "required": ["a", "c"],
            "type": "object",
            "properties": {
                "a": {"type": "string"},
                "c": {
                    "type": "array",
                    "items": {"type": "integer"}
                }
            }
        })

    def test_object_in_array(self):
        self.add_object([
            {"name": "Sir Lancelot of Camelot",
             "quest": "to seek the Holy Grail",
             "favorite colour": "blue"},
            {"name": "Sir Robin of Camelot",
             "quest": "to seek the Holy Grail",
             "capitol of Assyria": None}])
        self.assertResult({
            "type": "array",
            "items": {
                "type": "object",
                "required": ["name", "quest"],
                "properties": {
                    "quest": {"type": "string"},
                    "name": {"type": "string"},
                    "favorite colour": {"type": "string"},
                    "capitol of Assyria": {"type": "null"}
                }
            }
        })

    def test_three_deep(self):
        self.add_object({"matryoshka": {"design": {"principle": "FTW!"}}})
        self.assertResult({
            "type": "object",
            "required": ["matryoshka"],
            "properties": {
                "matryoshka": {
                    "type": "object",
                    "required": ["design"],
                    "properties": {
                        "design": {
                            "type": "object",
                            "required": ["principle"],
                            "properties": {
                                "principle": {"type": "string"}
                            }
                        }
                    }
                }
            }
        })



import unittest
from genson import SchemaBuilder
from genson.schema.strategies import (
    SchemaStrategy, Number, BASIC_SCHEMA_STRATEGIES)
from . import base


class MaxTenStrategy(Number):
    KEYWORDS = tuple(list(Number.KEYWORDS) + ['maximum'])

    def to_schema(self):
        schema = super().to_schema()
        schema['maximum'] = 10
        return schema


class FalseStrategy(SchemaStrategy):
    KEYWORDS = tuple(list(SchemaStrategy.KEYWORDS) + ['const'])

    @classmethod
    def match_schema(self, schema):
        return True

    @classmethod
    def match_object(self, obj):
        return True

    def to_schema(self):
        schema = super().to_schema()
        schema['type'] = 'boolean'
        schema['const'] = False
        return schema


class MaxTenSchemaBuilder(SchemaBuilder):
    EXTRA_STRATEGIES = (MaxTenStrategy,)


class FalseSchemaBuilder(SchemaBuilder):
    STRATEGIES = (FalseStrategy,)


class TestExtraStrategies(base.SchemaNodeTestCase):
    CLASS = MaxTenSchemaBuilder

    def test_add_object(self):
        self.add_object(5)
        self.assertResult({
            '$schema': 'http://json-schema.org/schema#',
            'type': 'integer',
            'maximum': 10})

    def test_add_schema(self):
        self.add_schema({'type': 'integer'})
        self.assertResult({
            '$schema': 'http://json-schema.org/schema#',
            'type': 'integer',
            'maximum': 10})


class DuplicateStrategiesSchemaBuilder(SchemaBuilder):
    """many duplicate strategies to exercise deduplication"""
    EXTRA_STRATEGIES = BASIC_SCHEMA_STRATEGIES * 100


class TestStrategyDeduplication(unittest.TestCase):

    def test_deduplicates_strategies(self):
        strategies = DuplicateStrategiesSchemaBuilder.STRATEGIES
        self.assertEqual(len(strategies), len(set(strategies)))

    def test_preserves_first_occurrence_order(self):
        self.assertEqual(DuplicateStrategiesSchemaBuilder.STRATEGIES,
                         tuple(BASIC_SCHEMA_STRATEGIES))

    def test_extra_strategies_keep_priority(self):
        # EXTRA_STRATEGIES stay ahead of the inherited strategies even
        # when they duplicate each other
        class DupExtra(SchemaBuilder):
            EXTRA_STRATEGIES = (MaxTenStrategy, MaxTenStrategy)

        self.assertIs(DupExtra.STRATEGIES[0], MaxTenStrategy)
        self.assertEqual(DupExtra.STRATEGIES.count(MaxTenStrategy), 1)


class TestClobberStrategies(base.SchemaNodeTestCase):
    CLASS = FalseSchemaBuilder

    def test_add_object(self):
        self.add_object("Any Norwegian Jarlsberger?")
        self.assertResult({
            '$schema': 'http://json-schema.org/schema#',
            'type': 'boolean',
            'const': False}, enforceUserContract=False)

    def test_add_schema(self):
        self.add_schema({'type': 'string'})
        self.assertResult({
            '$schema': 'http://json-schema.org/schema#',
            'type': 'boolean',
            'const': False}, enforceUserContract=False)



from . import base


class TestBasicTypes(base.SchemaNodeTestCase):

    def test_single_type(self):
        self.add_object("bacon")
        self.add_object("egg")
        self.add_object("spam")
        self.assertResult({"type": "string"})

    def test_redundant_integer_type(self):
        self.add_object(1)
        self.add_object(1.1)
        self.assertResult({"type": "number"})


class TestAnyOf(base.SchemaNodeTestCase):

    def test_simple(self):
        self.add_object("string")
        self.add_object(1.1)
        self.add_object(True)
        self.add_object(None)
        self.assertResult({"type": ["boolean", "null", "number", "string"]})

    def test_complex(self):
        self.add_object({})
        self.add_object([None])
        self.assertResult({"anyOf": [
            {"type": "object"},
            {"type": "array", "items": {"type": "null"}}
        ]})

    def test_simple_and_complex(self):
        self.add_object(None)
        self.add_object([None])
        self.assertResult({"anyOf": [
            {"type": "null"},
            {"type": "array", "items": {"type": "null"}}
        ]})


class TestArrayList(base.SchemaNodeTestCase):

    def setUp(self):
        base.SchemaNodeTestCase.setUp(self)

    def test_empty(self):
        self.add_object([])
        self.add_object([])

        self.assertResult({"type": "array"})

    def test_monotype(self):
        self.add_object(["spam", "spam", "spam", "eggs", "spam"])
        self.add_object(["spam", "bacon", "eggs", "spam"])

        self.assertResult({"type": "array", "items": {"type": "string"}})

    def test_multitype(self):
        self.add_object([1, "2", "3", None, False])
        self.add_object([1, 2, "3", False])

        self.assertObjectValidates([1, "2", 3, None, False])
        self.assertResult({
            "type": "array",
            "items": {
                "type": ["boolean", "integer", "null", "string"]}
        })

    def test_nested(self):
        self.add_object([
            ["surprise"],
            ["fear", "surprise"]
        ])
        self.add_object([
            ["fear", "surprise", "ruthless efficiency"],
            ["fear", "surprise", "ruthless efficiency",
             "an almost fanatical devotion to the Pope"]
        ])
        self.assertResult({
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "string"}}
        })


class TestArrayTuple(base.SchemaNodeTestCase):

    def test_empty(self):
        self.add_schema({"type": "array", "items": []})

        self.add_object([])
        self.add_object([])

        self.assertResult({"type": "array", "items": [{}]})

    def test_multitype(self):
        self.add_schema({"type": "array", "items": []})

        self.add_object([1, "2", "3", None, False])
        self.add_object([1, 2, "3", False])

        self.assertObjectDoesNotValidate([1, "2", 3, None, False])
        self.assertResult({
            "type": "array",
            "items": [
                {"type": "integer"},
                {"type": ["integer", "string"]},
                {"type": "string"},
                {"type": ["boolean", "null"]},
                {"type": "boolean"}]
        })

    def test_nested(self):
        self.add_schema(
            {"type": "array", "items": {"type": "array", "items": []}})

        self.add_object([
            ["surprise"],
            ["fear", "surprise"]
        ])
        self.add_object([
            ["fear", "surprise", "ruthless efficiency"],
            ["fear", "surprise", "ruthless efficiency",
             "an almost fanatical devotion to the Pope"]
        ])

        self.assertResult({
            "type": "array",
            "items": {
                "type": "array",
                "items": [
                    {"type": "string"},
                    {"type": "string"},
                    {"type": "string"},
                    {"type": "string"}
                ]
            }
        })



from .base import SchemaStrategy


class Enum(SchemaStrategy):
    """
    strategy for 'enum' schemas. Merges the 'enum' types into one set and then
    returns a schema with 'enum' set to the corresponding list.
    """

    KEYWORDS = ("enum",)

    def __init__(self, node_class):
        super().__init__(node_class)
        # Use set to easily merge 'enum's from different schemas.
        self._enum = set()
        # Apply different matching logic depending on whether schema exists or
        # not. Give the Enum strategy preference over other strategies. So
        # there's no way the Enum will get created unless it was explicitly
        # asked for by a schema, and this behavior is not order- or
        # type-dependent.
        self.match_object = self._instance_match_object

    @staticmethod
    def match_object(obj):
        # Exclude the strategy from the basic object matching. It must be
        # explicitly selected by a schema.
        return False

    def _instance_match_object(self, obj):
        # The strategy is selected by a schema. Give the Enum strategy
        # preference over all other strategies.
        return True

    @staticmethod
    def match_schema(schema):
        return "enum" in schema

    def add_schema(self, schema):
        super().add_schema(schema)
        if self._enum == set():
            self._enum = set(schema.get("enum"))
        elif "enum" in schema:
            self._enum.update(schema["enum"])

    def add_object(self, obj):
        super().add_object(obj)
        # Add only scalar types. Technically, the JSON-Schema spec allows
        # any type in an enum list, but using objects and lists is a very
        # rare use-case.
        if not isinstance(obj, (bool, str, int, float, type(None))):
            raise TypeError(f"Unsupported enum type of {type(obj)}. "
                            "Scalar type is expected.")
        # Convert to list to unify processing of string and other types.
        self._enum.update([obj])

    def to_schema(self):
        schema = super().to_schema()
        # Revert set back to list
        schema["enum"] = list(self._enum)
        return schema



import unittest
import json
import os
import sys
from subprocess import Popen, PIPE
from genson import SchemaBuilder

BASE_SCHEMA = {"$schema": SchemaBuilder.DEFAULT_URI}
FIXTURE_PATH = os.path.join(os.path.dirname(__file__), 'fixtures')
SHORT_USAGE = """\
usage: genson [-h] [--version] [-d DELIM] [-e ENCODING] [-i SPACES]
              [-s SCHEMA] [-$ SCHEMA_URI]
              ..."""


def fixture(filename):
    return os.path.join(FIXTURE_PATH, filename)


def stderr_message(message):
    return '{}\ngenson: error: {}\n'.format(SHORT_USAGE, message)


def run(args=tuple(), stdin_data=None, tty_stdin=False):
    """
    Run the ``genson`` executable as a subprocess and return
    (stdout, stderr). ``tty_stdin`` attaches stdin to a pseudo-TTY so
    the CLI treats the session as interactive; otherwise stdin is a
    pipe (with ``stdin_data`` written to it, if given).
    """
    full_args = [sys.executable, '-m', 'genson']
    full_args.extend(args)
    env = os.environ.copy()
    env['COLUMNS'] = '80'  # set width for deterministic text wrapping

    if tty_stdin:
        import pty
        master_fd, slave_fd = pty.openpty()
        stdin = slave_fd
    else:
        master_fd = slave_fd = None
        stdin = PIPE if stdin_data is not None else None

    genson_process = Popen(
        full_args, env=env, stdout=PIPE, stderr=PIPE, stdin=stdin)
    if stdin_data is not None:
        stdin_data = stdin_data.encode('utf-8')
    (stdout, stderr) = genson_process.communicate(stdin_data)
    genson_process.wait()
    if slave_fd is not None:
        os.close(slave_fd)
        os.close(master_fd)

    if isinstance(stdout, bytes):
        stdout = stdout.decode('utf-8')
    if isinstance(stderr, bytes):
        stderr = stderr.decode('utf-8')
    return (stdout, stderr)


class TestBasic(unittest.TestCase):

    def test_empty_input(self):
        (stdout, stderr) = run(stdin_data='')
        self.assertEqual(stderr, '')
        self.assertEqual(json.loads(stdout), BASE_SCHEMA)

    def test_empty_object_stdin(self):
        (stdout, stderr) = run(stdin_data='{}')
        self.assertEqual(stderr, '')
        self.assertEqual(
            json.loads(stdout),
            dict({"type": "object"}, **BASE_SCHEMA))

    def test_empty_object_file(self):
        (stdout, stderr) = run([fixture('empty.json')])
        self.assertEqual(stderr, '')
        self.assertEqual(
            json.loads(stdout),
            BASE_SCHEMA)

    def test_basic_schema_file(self):
        (stdout, stderr) = run(['-s', fixture('base_schema.json')])
        self.assertEqual(stderr, '')
        self.assertEqual(
            json.loads(stdout),
            BASE_SCHEMA)


class TestError(unittest.TestCase):
    maxDiff = 1000
    BAD_JSON_FILE = fixture('not_json.txt')
    BAD_JSON_MESSAGE = stderr_message(
        'invalid JSON in %s: Expecting value: line 1 column 1 (char 0)'
        % BAD_JSON_FILE)

    def assertEqualIgnoreWhitespace(
        self,
        first: str,
        second: str,
        msg: "str|None" = None,
    ):
        return self.assertEqual(
            " ".join(first.split()),
            " ".join(second.split()),
            msg=msg
        )

    @unittest.skipUnless(os.name == 'posix', 'pty requires a POSIX system')
    def test_no_input(self):
        # attach a pseudo-TTY so the CLI sees an interactive session
        # with no piped input rather than inheriting the runner's stdin
        (stdout, stderr) = run(tty_stdin=True)
        self.assertEqualIgnoreWhitespace(stderr, stderr_message(
            'nothing to do - no schemas or objects given'))
        self.assertEqual(stdout, '')

    def test_object_not_json(self):
        (stdout, stderr) = run([self.BAD_JSON_FILE])
        self.assertEqualIgnoreWhitespace(stderr, self.BAD_JSON_MESSAGE)
        self.assertEqual(stdout, '')

    def test_schema_not_json(self):
        (stdout, stderr) = run(['-s', self.BAD_JSON_FILE])
        self.assertEqualIgnoreWhitespace(stderr, self.BAD_JSON_MESSAGE)
        self.assertEqual(stdout, '')


class TestDelimiter(unittest.TestCase):

    def test_delim_newline(self):
        (stdout, stderr) = run(['-d', 'newline'],
                               stdin_data='{"hi":"there"}\n{"hi":5}')
        self.assertEqual(stderr, '')
        self.assertEqual(
            json.loads(stdout),
            dict({"required": ["hi"], "type": "object", "properties": {
                "hi": {"type": ["integer", "string"]}}}, **BASE_SCHEMA))

    def test_delim_auto_empty(self):
        (stdout, stderr) = run(['-d', ''], stdin_data='{"hi":"there"}{"hi":5}')
        self.assertEqual(stderr, '')
        self.assertEqual(
            json.loads(stdout),
            dict({"required": ["hi"], "type": "object", "properties": {
                "hi": {"type": ["integer", "string"]}}}, **BASE_SCHEMA))

    def test_delim_auto_whitespace(self):
        (stdout, stderr) = run(['-d', ''],
                               stdin_data='{"hi":"there"} \n\t{"hi":5}')
        self.assertEqual(stderr, '')
        self.assertEqual(
            json.loads(stdout),
            dict({"required": ["hi"], "type": "object", "properties": {
                "hi": {"type": ["integer", "string"]}}}, **BASE_SCHEMA))


class TestEncoding(unittest.TestCase):

    def test_encoding_unicode(self):
        (stdout, stderr) = run(
            ['-e', 'utf-8', fixture('utf-8.json')])
        self.assertEqual(stderr, '')
        self.assertEqual(
            json.loads(stdout),
            dict({"type": "string"}, **BASE_SCHEMA))

    def test_encoding_cp1252(self):
        (stdout, stderr) = run(
            ['-e', 'cp1252', fixture('cp1252.json')])
        self.assertEqual(stderr, '')
        self.assertEqual(
            json.loads(stdout),
            dict({"type": "string"}, **BASE_SCHEMA))



from .base import SchemaStrategy


class BaseArray(SchemaStrategy):
    """
    abstract array schema strategy
    """
    KEYWORDS = ('type', 'items')

    @staticmethod
    def match_object(obj):
        return isinstance(obj, list)

    def to_schema(self):
        schema = super().to_schema()
        schema['type'] = 'array'
        if self._items:
            schema['items'] = self.items_to_schema()
        return schema


class List(BaseArray):
    """
    strategy for list-style array schemas. This is the default
    strategy for arrays.
    """
    @staticmethod
    def match_schema(schema):
        return schema.get('type') == 'array' \
            and isinstance(schema.get('items', {}), dict)

    def __init__(self, node_class):
        super().__init__(node_class)
        self._items = node_class()

    def add_schema(self, schema):
        super().add_schema(schema)
        if 'items' in schema:
            self._items.add_schema(schema['items'])

    def add_object(self, obj):
        for item in obj:
            self._items.add_object(item)

    def items_to_schema(self):
        return self._items.to_schema()


class Tuple(BaseArray):
    """
    strategy for tuple-style array schemas. These will always have
    an items key to preserve the fact that it's a tuple.
    """
    @staticmethod
    def match_schema(schema):
        return schema.get('type') == 'array' \
            and isinstance(schema.get('items'), list)

    def __init__(self, node_class):
        super().__init__(node_class)
        self._items = [node_class()]

    def add_schema(self, schema):
        super().add_schema(schema)
        if 'items' in schema:
            self._add(schema['items'], 'add_schema')

    def add_object(self, obj):
        self._add(obj, 'add_object')

    def _add(self, items, func):
        while len(self._items) < len(items):
            self._items.append(self.node_class())

        for subschema, item in zip(self._items, items):
            getattr(subschema, func)(item)

    def items_to_schema(self):
        return [item.to_schema() for item in self._items]



from . import base


class TestSeedTuple(base.SchemaNodeTestCase):

    def test_tuple(self):
        self.add_schema({'type': 'array', 'items': []})
        self.add_object([None])
        self.assertResult({'type': 'array', 'items': [{'type': 'null'}]})

    def test_empty_tuple(self):
        # an empty tuple seed is padded to one empty item schema; a
        # zero-length `items` array would be invalid under every draft
        # of the meta-schema (see the discussion on issue #90)
        self.add_schema({'type': 'array', 'items': []})
        self.assertResult({'type': 'array', 'items': [{}]})

    def test_empty_tuple_merge(self):
        self.add_schema({'type': 'array', 'items': []})
        self.add_schema({'type': 'array', 'items': []})
        self.assertResult({'type': 'array', 'items': [{}]})


class TestPatternProperties(base.SchemaNodeTestCase):

    def test_single_pattern(self):
        self.add_schema({'type': 'object', 'patternProperties': {
            r'^\d$': None}})
        self.add_object({'0': 0, '1': 1, '2': 2})
        self.assertResult({'type': 'object', 'patternProperties': {
            r'^\d$': {'type': 'integer'}}})

    def test_multi_pattern(self):
        self.add_schema({'type': 'object', 'patternProperties': {
            r'^\d$': None,
            r'^[a-z]$': None}})
        self.add_object({'0': 0, '1': 1, 'a': True, 'b': False})
        self.assertResult({'type': 'object', 'patternProperties': {
            r'^\d$': {'type': 'integer'},
            r'^[a-z]$': {'type': 'boolean'}}})

    def test_multi_pattern_multi_object(self):
        self.add_schema({'type': 'object', 'patternProperties': {
            r'^\d$': None,
            r'^[a-z]$': None}})
        self.add_object({'0': 0})
        self.add_object({'1': 1})
        self.add_object({'a': True})
        self.add_object({'b': False})
        self.assertResult({'type': 'object', 'patternProperties': {
            r'^\d$': {'type': 'integer'},
            r'^[a-z]$': {'type': 'boolean'}}})

    def test_existing_schema(self):
        self.add_schema({'type': 'object', 'patternProperties': {
            r'^\d$': {'type': 'boolean'}}})
        self.add_object({'0': 0, '1': 1, '2': 2})
        self.assertResult({'type': 'object', 'patternProperties': {
            r'^\d$': {'type': ['boolean', 'integer']}}})

    def test_prefers_existing_properties(self):
        self.add_schema({'type': 'object',
                         'properties': {'0': None},
                         'patternProperties': {r'^\d$': None}})
        self.add_object({'0': 0, '1': 1, '2': 2})
        self.assertResult({'type': 'object',
                           'properties': {'0': {'type': 'integer'}},
                           'patternProperties': {r'^\d$': {'type': 'integer'}},
                           'required': ['0']})

    def test_keeps_unrecognized_properties(self):
        self.add_schema({'type': 'object',
                         'patternProperties': {r'^\d$': None}})
        self.add_object({'0': 0, '1': 1, '2': 2, 'a': True})
        self.assertResult({'type': 'object',
                           'properties': {'a': {'type': 'boolean'}},
                           'patternProperties': {r'^\d$': {'type': 'integer'}},
                           'required': ['a']})


class TestEnum(base.SchemaNodeTestCase):

    def test_enum_scalar_string(self):
        self.add_schema({"enum": []})
        self.add_object("1")
        self.assertResult({"enum": ["1"]})

    def test_enum_scalar_list(self):
        self.add_schema({"enum": ["1", 2]})
        self.add_object("34")
        self.add_object(5)
        self.add_object(6.7)
        self.add_object(False)
        self.add_object(None)
        self.assertResult(
            {"enum": ["1", 2, "34", 5, 6.7, False, None]},
            enforceUserContract=False,
            ignore_order=True,
        )

    def test_enum_repeated_item(self):
        self.add_schema({"enum": []})
        self.add_object(1)
        self.add_object(1)
        self.add_object(2)
        self.add_object(1)
        self.assertResult({"enum": [1, 2]}, ignore_order=True)



from .base import SchemaStrategy, TypedSchemaStrategy


class Typeless(SchemaStrategy):
    """
    schema strategy for schemas with no type. This is only used when
    there is no other active strategy, and it will be merged into the
    first typed strategy that gets added.
    """

    @classmethod
    def match_schema(cls, schema):
        return 'type' not in schema

    @classmethod
    def match_object(cls, obj):
        return False


class Null(TypedSchemaStrategy):
    """
    strategy for null schemas
    """
    JS_TYPE = 'null'
    PYTHON_TYPE = type(None)


class Boolean(TypedSchemaStrategy):
    """
    strategy for boolean schemas
    """
    JS_TYPE = 'boolean'
    PYTHON_TYPE = bool


class String(TypedSchemaStrategy):
    """
    strategy for string schemas - works for ascii and unicode strings
    """
    JS_TYPE = 'string'
    PYTHON_TYPE = str


class Number(SchemaStrategy):
    """
    strategy for integer and number schemas. It automatically
    converts from `integer` to `number` when a float object or a
    number schema is added
    """
    JS_TYPES = ('integer', 'number')
    PYTHON_TYPES = (int, float)

    @classmethod
    def match_schema(cls, schema):
        return schema.get('type') in cls.JS_TYPES

    @classmethod
    def match_object(cls, obj):
        # cannot use isinstance() because boolean is a subtype of int
        return type(obj) in cls.PYTHON_TYPES

    def __init__(self, node_class):
        super().__init__(node_class)
        self._type = 'integer'

    def add_schema(self, schema):
        super().add_schema(schema)
        if schema.get('type') == 'number':
            self._type = 'number'

    def add_object(self, obj):
        super().add_object(obj)
        if isinstance(obj, float):
            self._type = 'number'

    def to_schema(self):
        schema = super().to_schema()
        schema['type'] = self._type
        return schema



from .base import (
    SchemaStrategy,
    TypedSchemaStrategy
)
from .scalar import (
    Typeless,
    Null,
    Boolean,
    Number,
    String
)
from .array import List, Tuple
from .enum import Enum
from .object import Object

BASIC_SCHEMA_STRATEGIES = (
    Null,
    Boolean,
    Number,
    String,
    List,
    Tuple,
    Object,
    Enum,
)

__all__ = (
    'SchemaStrategy',
    'TypedSchemaStrategy',
    'Null',
    'Boolean',
    'Number',
    'String',
    'List',
    'Tuple',
    'Object',
    'Typeless',
    'Enum',
    'BASIC_SCHEMA_STRATEGIES'
)



def sort_lists_in_schema(schema):
    """
    Recursively traverses a schema and sorts any lists it finds.
    The sorting is stable and works on lists with mixed types.
    """
    if isinstance(schema, dict):
        for key, value in schema.items():
            schema[key] = sort_lists_in_schema(value)
        return schema
    elif isinstance(schema, list):
        new_list = [sort_lists_in_schema(item) for item in schema]
        try:
            return sorted(new_list)
        except TypeError:
            # This handles mixed types by creating a tuple of the type
            # name and the value. This ensures that types are grouped
            # together and then sorted.
            # It's safe to use str() here since only malformed custom
            # objects could cause it to fail.
            return sorted(new_list, key=lambda x: (type(x).__name__, str(x)))
    return schema



from .schema.builder import SchemaBuilder, Schema
from .schema.node import SchemaNode, SchemaGenerationError
from .schema.strategies.base import SchemaStrategy, TypedSchemaStrategy

__version__ = '1.4.0'
__all__ = [
    'SchemaBuilder',
    'SchemaNode',
    'SchemaGenerationError',
    'Schema',
    'SchemaStrategy',
    'TypedSchemaStrategy']



from . import base


class TestType(base.SchemaNodeTestCase):

    def test_single_type(self):
        schema = {'type': 'string'}
        self.add_schema(schema)
        self.assertResult(schema)

    def test_single_type_unicode(self):
        schema = {u'type': u'string'}
        self.add_schema(schema)
        self.assertResult(schema)

    def test_typeless(self):
        schema = {}
        self.add_schema(schema)
        self.assertResult(schema)

    def test_array_type_no_items(self):
        schema = {'type': 'array'}
        self.add_schema(schema)
        self.assertResult(schema)


class TestAnyOf(base.SchemaNodeTestCase):

    def test_multi_type(self):
        schema = {'type': ['boolean', 'null', 'number', 'string']}
        self.add_schema(schema)
        self.assertResult(schema)

    def test_multi_type_with_extra_keywords(self):
        schema = {'type': ['boolean', 'null', 'number', 'string'],
                  'title': 'this will be duplicated'}
        self.add_schema(schema)
        self.assertResult({'anyOf': [
            {'type': 'boolean', 'title': 'this will be duplicated'},
            {'type': 'null', 'title': 'this will be duplicated'},
            {'type': 'number', 'title': 'this will be duplicated'},
            {'type': 'string', 'title': 'this will be duplicated'}
        ]})

    def test_anyof(self):
        schema = {"anyOf": [
            {"type": "null"},
            {"type": "boolean", "title": "Gruyere"}
        ]}
        self.add_schema(schema)
        self.assertResult(schema)

    def test_recursive(self):
        schema = {"anyOf": [
            {"type": ["integer", "string"]},
            {"anyOf": [
                {"type": "null"},
                {"type": "boolean", "title": "Gruyere"}
            ]}
        ]}
        self.add_schema(schema)
        # recursive anyOf will be flattened
        self.assertResult({"anyOf": [
            {"type": ["integer", "null", "string"]},
            {"type": "boolean", "title": "Gruyere"}
        ]})


class TestRequired(base.SchemaNodeTestCase):

    def test_preserves_empty_required(self):
        schema = {'type': 'object', 'required': []}
        self.add_schema(schema)
        self.assertResult(schema)


class TestPreserveExtraKeywords(base.SchemaNodeTestCase):

    def test_basic_type(self):
        schema = {'type': 'boolean', 'const': False, 'myKeyword': True}
        self.add_schema(schema)
        self.assertResult(schema)

    def test_number(self):
        schema = {'type': 'number', 'const': 5, 'myKeyword': True}
        self.add_schema(schema)
        self.assertResult(schema)

    def test_list(self):
        schema = {'type': 'array', 'items': {"type": "null"},
                  'const': [], 'myKeyword': True}
        self.add_schema(schema)
        self.assertResult(schema)

    def test_tuple(self):
        schema = {'type': 'array', 'items': [{"type": "null"}],
                  'const': [], 'myKeyword': True}
        self.add_schema(schema)
        self.assertResult(schema)

    def test_object(self):
        schema = {'type': 'object', 'const': {}, 'myKeyword': True}
        self.add_schema(schema)
        self.assertResult(schema)

    def test_typeless(self):
        schema = {'const': 5, 'myKeyword': True}
        self.add_schema(schema)
        self.assertResult(schema)



from . import base


class TestType(base.SchemaNodeTestCase):

    def test_single_type(self):
        schema = {'type': 'null'}
        self.add_schema(schema)
        self.add_schema(schema)
        self.assertResult(schema)

    def test_single_type_unicode(self):
        schema = {u'type': u'string'}
        self.add_schema(schema)
        self.assertResult(schema)

    def test_redundant_integer_type(self):
        self.add_schema({'type': 'integer'})
        self.add_schema({'type': 'number'})
        self.assertResult({'type': 'number'})

    def test_typeless(self):
        schema1 = {"title": "ambiguous schema"}
        schema2 = {"grail": "We've already got one"}
        result = dict(schema1)
        result.update(schema2)
        self.add_schema(schema1)
        self.add_schema(schema2)
        self.assertResult(result)

    def test_typeless_incorporated(self):
        schema1 = {"title": "Gruyere"}
        schema2 = {"type": "boolean"}
        self.add_schema(schema1)
        self.add_schema(schema2)
        self.assertResult({"type": "boolean", "title": "Gruyere"})

    def test_typeless_instantly_incorporated(self):
        schema1 = {"type": "boolean"}
        schema2 = {"title": "Gruyere"}
        self.add_schema(schema1)
        self.add_schema(schema2)
        self.assertResult({"type": "boolean", "title": "Gruyere"})


class TestRequired(base.SchemaNodeTestCase):

    def test_combines(self):
        schema1 = {"type": "object", "required": [
            "series of statements", "definite proposition"]}
        schema2 = {"type": "object", "required": ["definite proposition"]}
        self.add_schema(schema1)
        self.add_schema(schema2)
        self.assertResult({"type": "object", "required": [
            "definite proposition"]})

    def test_ignores_missing(self):
        schema1 = {"type": "object"}
        schema2 = {"type": "object", "required": ["definite proposition"]}
        self.add_schema(schema1)
        self.add_schema(schema2)
        self.assertResult({"type": "object", "required": [
            "definite proposition"]})

    def test_omits_all_missing(self):
        schema1 = {"type": "object", "properties": {"spam": {}}}
        schema2 = {"type": "object", "properties": {"eggs": {}}}
        self.add_schema(schema1)
        self.add_schema(schema2)
        self.assertResult(
            {"type": "object", "properties": {"spam": {}, "eggs": {}}})

    def test_maintains_empty(self):
        seed = {"required": []}
        schema1 = {"type": "object", "required": ["series of statements"]}
        schema2 = {"type": "object", "required": ["definite proposition"]}
        self.add_schema(seed)
        self.add_schema(schema1)
        self.add_schema(schema2)
        self.assertResult({"type": "object", "required": []})


class TestAnyOf(base.SchemaNodeTestCase):

    def test_multi_type(self):
        self.add_schema({'type': 'boolean'})
        self.add_schema({'type': 'null'})
        self.add_schema({'type': 'string'})
        self.assertResult({'type': ['boolean', 'null', 'string']})

    def test_anyof_generated(self):
        schema1 = {"type": "null", "title": "African or European Swallow?"}
        schema2 = {"type": "boolean", "title": "Gruyere"}
        self.add_schema(schema1)
        self.add_schema(schema2)
        self.assertResult({"anyOf": [
            schema1,
            schema2
        ]})

    def test_anyof_seeded(self):
        schema1 = {"type": "null", "title": "African or European Swallow?"}
        schema2 = {"type": "boolean", "title": "Gruyere"}
        self.add_schema({"anyOf": [
            {"type": "null"},
            schema2
        ]})
        self.add_schema(schema1)
        self.assertResult({"anyOf": [
            schema1,
            schema2
        ]})

    def test_list_plus_tuple(self):
        schema1 = {"type": "array", "items": {"type": "null"}}
        schema2 = {"type": "array", "items": [{"type": "null"}]}
        self.add_schema(schema1)
        self.add_schema(schema2)
        self.assertResult({"anyOf": [
            schema1,
            schema2
        ]})

    def test_multi_type_and_anyof(self):
        schema1 = {'type': ['boolean', 'null', 'string']}
        schema2 = {"type": "boolean", "title": "Gruyere"}
        self.add_schema(schema1)
        self.add_schema(schema2)
        self.assertResult({"anyOf": [
            {'type': ['null', 'string']},
            schema2
        ]})



import argparse
import sys
import re
import json
from . import SchemaBuilder, __version__


class CLI:
    def __init__(self, prog=None):
        self._make_parser(prog)
        self._prepare_args()
        self.builder = SchemaBuilder(schema_uri=self.args.schema_uri)

    def run(self):
        if not self.args.schema and not self.args.object:
            self.fail('nothing to do - no schemas or objects given')
        self.add_schemas()
        self.add_objects()
        self.print_output()

    def add_schemas(self):
        for fp in self.args.schema:
            self._call_with_json_from_fp(self.builder.add_schema, fp)
            fp.close()

    def add_objects(self):
        for fp in self.args.object:
            self._call_with_json_from_fp(self.builder.add_object, fp)
            fp.close()

    def print_output(self):
        print(self.builder.to_json(indent=self.args.indent))

    def fail(self, message):
        self.parser.error(message)

    def _make_parser(self, prog=None):
        file_type = argparse.FileType('r', encoding=self._get_encoding())

        self.parser = argparse.ArgumentParser(
            add_help=False,
            prog=prog,
            description="""Generate one, unified JSON Schema from one or more
            JSON objects and/or JSON Schemas. Compatible with JSON-Schema Draft
            4 and above.""")

        self.parser.add_argument(
            '-h', '--help', action='help', default=argparse.SUPPRESS,
            help='Show this help message and exit.')
        self.parser.add_argument(
            '--version', action='version', default=argparse.SUPPRESS,
            version='%(prog)s {}'.format(__version__),
            help='Show version number and exit.')
        self.parser.add_argument(
            '-d', '--delimiter', metavar='DELIM',
            help="""Set a delimiter. Use this option if the input files
            contain multiple JSON objects/schemas. You can pass any string. A
            few cases ('newline', 'tab', 'space') will get converted to a
            whitespace character. If this option is omitted, the parser will
            try to auto-detect boundaries.""")
        self.parser.add_argument(
            '-e', '--encoding', type=str, metavar='ENCODING',
            help="""Use ENCODING instead of the default system encoding
            when reading files. ENCODING must be a valid codec name or
            alias.""")
        self.parser.add_argument(
            '-i', '--indent', type=int, metavar='SPACES',
            help="""Pretty-print the output, indenting SPACES spaces.""")
        self.parser.add_argument(
            '-s', '--schema', action='append', default=[], type=file_type,
            help="""File containing a JSON Schema (can be specified multiple
            times to merge schemas).""")
        self.parser.add_argument(
            '-$', '--schema-uri', metavar='SCHEMA_URI', dest='schema_uri',
            default=SchemaBuilder.DEFAULT_URI,
            help="""The value of the '$schema' keyword (defaults to {default!r}
            or can be specified in a schema with the -s option). If {null!r} is
            passed, the "$schema" keyword will not be included in the
            result.""".format(default=SchemaBuilder.DEFAULT_URI,
                              null=SchemaBuilder.NULL_URI))
        self.parser.add_argument(
            'object', nargs=argparse.REMAINDER, type=file_type,
            help="""Files containing JSON objects (defaults to stdin if no
            arguments are passed).""")

    def _get_encoding(self):
        """
        use separate arg parser to grab encoding argument before
        defining FileType args
        """
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('-e', '--encoding', type=str)
        args, _ = parser.parse_known_args()
        return args.encoding

    def _prepare_args(self):
        self.args = self.parser.parse_args()
        self._prepare_delimiter()

        # default to stdin if no objects or schemas
        if not self.args.object and not sys.stdin.isatty():
            self.args.object.append(sys.stdin)

    def _prepare_delimiter(self):
        """
        manage special conversions for difficult bash characters
        """
        if self.args.delimiter == 'newline':
            self.args.delimiter = '\n'
        elif self.args.delimiter == 'tab':
            self.args.delimiter = '\t'
        elif self.args.delimiter == 'space':
            self.args.delimiter = ' '

    def _call_with_json_from_fp(self, method, fp):
        for json_string in self._get_json_strings(fp.read().strip()):
            try:
                json_obj = json.loads(json_string)
            except json.JSONDecodeError as err:
                self.fail('invalid JSON in {}: {}'.format(fp.name, err))
            method(json_obj)

    def _get_json_strings(self, raw_text):
        if self.args.delimiter is None or self.args.delimiter == '':
            json_strings = self._detect_json_strings(raw_text)
        else:
            json_strings = raw_text.split(self.args.delimiter)

        # sanitize data before returning
        return [string.strip() for string in json_strings if string.strip()]

    @staticmethod
    def _detect_json_strings(raw_text):
        """
        Use regex with lookaround to spot the boundaries between JSON
        objects. Unfortunately, it has to match *something*, so at least
        one character must be removed and replaced.
        """
        strings = re.split(r'}\s*(?={)', raw_text)

        # put back the stripped character
        json_strings = [string + '}' for string in strings[:-1]]

        # the last one doesn't need to be modified
        json_strings.append(strings[-1])

        return json_strings


def main():
    CLI().run()


if __name__ == "__main__":
    CLI('genson').run()



from . import base
from genson import SchemaBuilder

SCHEMA_URI = 'https://json-schema.org/draft/2020-12/schema'


class TestParams(base.SchemaBuilderTestCase):

    def test_uri(self):
        self.builder = SchemaBuilder(schema_uri=SCHEMA_URI)
        self.assertResult({"$schema": SCHEMA_URI})

    def test_null_uri(self):
        self.builder = SchemaBuilder(schema_uri=None)
        self.assertResult({})


class TestMethods(base.SchemaBuilderTestCase):

    def test_add_schema(self):
        self.add_schema({"type": "null"})
        self.assertResult({
            "$schema": SchemaBuilder.DEFAULT_URI,
            "type": "null"})

    def test_add_object(self):
        self.add_object(None)
        self.assertResult({
            "$schema": SchemaBuilder.DEFAULT_URI,
            "type": "null"})

    def test_to_json(self):
        self.assertEqual(
            self.builder.to_json(),
            '{"$schema": "%s"}' % SchemaBuilder.DEFAULT_URI)

    def test_add_schema_with_uri_default(self):
        self.add_schema({"$schema": SCHEMA_URI, "type": "null"})
        self.assertResult({"$schema": SCHEMA_URI, "type": "null"})

    def test_add_schema_with_uri_not_defuult(self):
        self.builder = SchemaBuilder(schema_uri=SCHEMA_URI)
        self.add_schema({"$schema": 'BAD_URI', "type": "null"})
        self.assertResult({"$schema": SCHEMA_URI, "type": "null"})

    def test_empty_falsy(self):
        self.assertIs(bool(self.builder), False)

    def test_full_truty(self):
        self.add_object(None)
        self.assertIs(bool(self.builder), True)


class TestInteraction(base.SchemaBuilderTestCase):

    def test_add_other(self):
        other = SchemaBuilder(schema_uri=SCHEMA_URI)
        other.add_object(1)
        self.add_object('one')
        self.add_schema(other)
        self.assertResult({
            "$schema": SCHEMA_URI,
            "type": ["integer", "string"]})

    def test_add_other_no_uri_overwrite(self):
        other = SchemaBuilder()
        other.add_object(1)
        self.add_object('one')
        self.add_schema(other)
        self.add_schema({'$schema': SCHEMA_URI})
        self.assertResult({
            "$schema": SCHEMA_URI,
            "type": ["integer", "string"]})

    def test_eq(self):
        b1 = SchemaBuilder()
        b1.add_object(1)
        b2 = SchemaBuilder()
        b2.add_object(1)
        self.assertEqual(b1, b2)

    def test_ne(self):
        b1 = SchemaBuilder()
        b1.add_object(1)
        b2 = SchemaBuilder()
        b2.add_object('one')
        self.assertNotEqual(b1, b2)

    def test_eq_after_serialization(self):
        b1 = SchemaBuilder()
        b1.add_object({"bar": 10, "foo": 20})
        b2 = SchemaBuilder()
        b2.add_schema(b1)
        self.assertEqual(b1, b2)

    def test_eq_empty_required(self):
        b1 = SchemaBuilder()
        b1.add_schema({
            "type": "object",
            "properties": {
                "bar": {"type": "integer"},
                "foo": {"type": "integer"}},
            "required": []})
        b2 = SchemaBuilder()
        b2.add_schema(b1)
        self.assertEqual(b1, b2)



from .strategies import BASIC_SCHEMA_STRATEGIES, Typeless


class SchemaGenerationError(RuntimeError):
    pass


class SchemaNode:
    """
    Basic schema generator class. SchemaNode objects can be loaded
    up with existing schemas and objects before being serialized.
    """
    STRATEGIES = BASIC_SCHEMA_STRATEGIES

    def __init__(self):
        self._active_strategies = []

    def add_schema(self, schema):
        """
        Merges in an existing schema.

        arguments:
        * `schema` (required - `dict` or `SchemaNode`):
          an existing JSON Schema to merge.
        """

        # serialize instances of SchemaNode before parsing
        if isinstance(schema, SchemaNode):
            schema = schema.to_schema()

        for subschema in self._get_subschemas(schema):
            # delegate to SchemaType object
            active_strategy = self._get_strategy_for_schema(subschema)
            active_strategy.add_schema(subschema)

        # return self for easy method chaining
        return self

    def add_object(self, obj):
        """
        Modify the schema to accommodate an object.

        arguments:
        * `obj` (required - `dict`):
          a JSON object to use in generating the schema.
        """

        # delegate to SchemaType object
        active_strategy = self._get_strategy_for_object(obj)
        active_strategy.add_object(obj)

        # return self for easy method chaining
        return self

    def to_schema(self):
        """
        Convert the current schema to a `dict`.
        """
        types = set()
        generated_schemas = []
        for active_strategy in self._active_strategies:
            generated_schema = active_strategy.to_schema()
            if len(generated_schema) == 1 and 'type' in generated_schema:
                types.add(generated_schema['type'])
            else:
                generated_schemas.append(generated_schema)

        if types:
            if len(types) == 1:
                (types,) = types
            else:
                types = sorted(types)
            generated_schemas = [{'type': types}] + generated_schemas
        if len(generated_schemas) == 1:
            (result_schema,) = generated_schemas
        elif generated_schemas:
            result_schema = {'anyOf': generated_schemas}
        else:
            result_schema = {}

        return result_schema

    def __len__(self):
        return len(self._active_strategies)

    def __eq__(self, other):
        """ Required for SchemaBuilder.__eq__ to work properly """
        return (isinstance(other, self.__class__)
                and self.__dict__ == other.__dict__)

    # private methods

    def _get_subschemas(self, schema):
        if 'anyOf' in schema:
            return [subschema for anyof in schema['anyOf']
                    for subschema in self._get_subschemas(anyof)]
        elif isinstance(schema.get('type'), list):
            other_keys = dict(schema)
            del other_keys['type']
            return [dict(type=tipe, **other_keys) for tipe in schema['type']]
        else:
            return [schema]

    def _get_strategy_for_schema(self, schema):
        return self._get_strategy_for_('schema', schema)

    def _get_strategy_for_object(self, obj):
        return self._get_strategy_for_('object', obj)

    def _get_strategy_for_(self, kind, schema_or_obj):
        # check existing types
        for active_strategy in self._active_strategies:
            if getattr(active_strategy, 'match_' + kind)(schema_or_obj):
                return active_strategy

        # check all potential types
        for strategy in self.STRATEGIES:
            if getattr(strategy, 'match_' + kind)(schema_or_obj):
                active_strategy = strategy(self.__class__)

                # incorporate typeless strategy if it exists
                if self._active_strategies and \
                        isinstance(self._active_strategies[-1], Typeless):
                    typeless = self._active_strategies.pop()
                    active_strategy.add_schema(typeless.to_schema())

                self._active_strategies.append(active_strategy)
                return active_strategy

        # no match found, if typeless add to first strategy
        if kind == 'schema' and Typeless.match_schema(schema_or_obj):
            if not self._active_strategies:
                self._active_strategies.append(Typeless(self.__class__))
            active_strategy = self._active_strategies[0]
            return active_strategy

        # no match found, raise an error
        raise SchemaGenerationError(
            'Could not find matching schema type for {0}: {1!r}'.format(
                kind, schema_or_obj))



from . import base
from genson import Schema, SchemaGenerationError


class TestMisuse(base.SchemaBuilderTestCase):

    def test_schema_with_bad_type_error(self):
        with self.assertRaises(SchemaGenerationError):
            self.add_schema({'type': 'african swallow'})

    def test_to_dict_pending_deprecation_warning(self):
        with self.assertWarns(PendingDeprecationWarning):
            builder = Schema()
        with self.assertWarns(PendingDeprecationWarning):
            builder.add_object('I fart in your general direction!')
            builder.to_dict()

    def test_recurse_deprecation_warning(self):
        with self.assertWarns(DeprecationWarning):
            builder = Schema()
            builder.add_object('Go away or I shall taunt you a second time!')
            builder.to_dict(recurse=True)

    def test_incompatible_schema_warning(self):
        with self.assertWarns(UserWarning):
            self.add_schema({'type': 'string', 'length': 5})
            self.add_schema({'type': 'string', 'length': 7})



import sys
import unittest
import jsonschema
from genson import SchemaNode, SchemaBuilder
from .sort import sort_lists_in_schema

PYTHON_VERSION = sys.version[:sys.version.find(' ')]


class BaseTestCase(unittest.TestCase):

    def setUp(self):
        self.builder = self.CLASS()
        self._objects = []
        self._schemas = []

    def set_schema_options(self, **options):
        self.builder = SchemaNode(**options)

    def add_object(self, obj):
        self.builder.add_object(obj)
        self._objects.append(obj)

    def add_schema(self, schema):
        self.builder.add_schema(schema)
        self._schemas.append(schema)

    def assertObjectValidates(self, obj):
        jsonschema.Draft7Validator(self.builder.to_schema()).validate(obj)

    def assertObjectDoesNotValidate(self, obj):
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            jsonschema.Draft7Validator(self.builder.to_schema()).validate(obj)

    def assertResult(self, expected, enforceUserContract=True,
                     ignore_order=False):
        actual = self.builder.to_schema()
        if ignore_order:
            sort_lists_in_schema(actual)
            sort_lists_in_schema(expected)
        self.assertEqual(
            expected, actual,
            'Generated schema (below) does not match expected (above)')
        if enforceUserContract:
            self.assertUserContract()

    def assertUserContract(self):
        self._assertSchemaIsValid()
        self._assertComponentObjectsValidate()

    def _assertSchemaIsValid(self):
        jsonschema.Draft7Validator.check_schema(self.builder.to_schema())

    def _assertComponentObjectsValidate(self):
        compiled_schema = self.builder.to_schema()
        for obj in self._objects:
            jsonschema.Draft7Validator(compiled_schema).validate(obj)


class SchemaNodeTestCase(BaseTestCase):
    CLASS = SchemaNode


class SchemaBuilderTestCase(BaseTestCase):
    CLASS = SchemaBuilder



from collections import defaultdict
from re import search
from .base import SchemaStrategy


class Object(SchemaStrategy):
    """
    object schema strategy
    """
    KEYWORDS = ('type', 'properties', 'patternProperties', 'required')

    @staticmethod
    def match_schema(schema):
        return schema.get('type') == 'object'

    @staticmethod
    def match_object(obj):
        return isinstance(obj, dict)

    def __init__(self, node_class):
        super().__init__(node_class)

        self._properties = defaultdict(node_class)
        self._pattern_properties = defaultdict(node_class)
        self._required = None
        self._include_empty_required = False

    def add_schema(self, schema):
        super().add_schema(schema)
        if 'properties' in schema:
            for prop, subschema in schema['properties'].items():
                subnode = self._properties[prop]
                if subschema is not None:
                    subnode.add_schema(subschema)
        if 'patternProperties' in schema:
            for pattern, subschema in schema['patternProperties'].items():
                subnode = self._pattern_properties[pattern]
                if subschema is not None:
                    subnode.add_schema(subschema)
        if 'required' in schema:
            required = set(schema['required'])
            if not required:
                self._include_empty_required = True
            if self._required is None:
                self._required = required
            else:
                self._required &= required

    def add_object(self, obj):
        properties = set()
        for prop, subobj in obj.items():
            pattern = None

            if prop not in self._properties:
                pattern = self._matching_pattern(prop)

            if pattern is not None:
                self._pattern_properties[pattern].add_object(subobj)
            else:
                properties.add(prop)
                self._properties[prop].add_object(subobj)

        if self._required is None:
            self._required = properties
        else:
            self._required &= properties

    def _matching_pattern(self, prop):
        for pattern in self._pattern_properties.keys():
            if search(pattern, prop):
                return pattern

    def to_schema(self):
        schema = super().to_schema()
        schema['type'] = 'object'
        if self._properties:
            schema['properties'] = self._properties_to_schema(
                self._properties)
        if self._pattern_properties:
            schema['patternProperties'] = self._properties_to_schema(
                self._pattern_properties)
        if self._required or self._include_empty_required:
            schema['required'] = sorted(self._required)
        return schema

    def _properties_to_schema(self, properties):
        schema_properties = {}
        for prop, schema_node in properties.items():
            schema_properties[prop] = schema_node.to_schema()
        return schema_properties
